from threading import Thread, Lock, Event
from copy import deepcopy, copy
from cherrypy import engine
import time, datetime
import cPickle
import sys
import random
from GuiUtils import _logerr, _logwarn, _loginfo


class OverviewCache(Thread):
  """
  Thread-watched cache of data for prodmon (or anything else)
  All non _-prefixed methods should be thread-safe for external calling
  """
  exthook = "OverviewCache"
  
  def __init__(self,gui,interval=3,sizelimit=100000000,itemlimit=1000):
    """
    Create a new ProdmonCache. Note that it is not started until the .start() method is called.
    
    @param interval: interval in seconds between the main loop running (3)
    @param cachesize: maximum size in bytes before old entries start getting culled. Note that this is done *imprecisely* and will not be strictly adhered to! (100MB)
    """
    _loginfo("CACHE: Creating new Overview cache interval=%s sec, sizelimit=%s bytes, itemlimit=%s"%(interval,sizelimit,itemlimit))
    
    Thread.__init__(self,name="OverviewCache")
    
    self.lock = Lock()
    self.stopme = False
    self.cachesize = 0
    self.cache = {} # objkey: tuple (expiry, size, data)
    self.interval = interval
    self.sizelimit = sizelimit
    self.itemlimit = itemlimit
    self.cacheevents = {}
    
    engine.subscribe('stop',self.stop)
    self.start()
    
    
  def stop(self):
    """
    When called, causes the thread to exit, the next time the main loop runs.
    """
    #_loginfo("CACHE: Trying to stop cache")
    self.lock.acquire()
    try:
      self.stopme = True
    finally:
      self.lock.release()
    
  def run(self):
    while True:
      self.lock.acquire()
      #_loginfo("CACHE: Main loop, size=%s, items=%s"%(self.cachesize,len(self.cache)))
      try:
        if self.stopme:
          return
        keys_to_delete=[]
        now = time.time()
        for key in self.cache:
          if self.cache[key][0]<now:
            keys_to_delete.append(key)
        for key in keys_to_delete:
          #_loginfo("CACHE: Deleting expired key: %s"%key)
          self._remove(key)
          
          #if there is an associated key
          if key in self.cacheevents:
            event = self.cacheevents[key]
            if not event.isSet():
              event.set()
            del self.cacheevents[key]
        
 
      
      finally:
        self.lock.release()
      time.sleep(self.interval)
  
  def _get_oldest(self):
    #print "get oldest"
    t = sys.maxint
    oldest = None
    for key in self.cache:
      if self.cache[key][0]<t:
        t = self.cache[key][0]
        oldest = key
    return oldest
      
  def _cull_number(self):
    #print "cull_number"
    #_loginfo("CACHE: Culling excessive item number (%s)"%len(self.cache))
    while len(self.cache)>self.itemlimit:
      self._remove(self._get_oldest())
    #print "done"
      
  def _cull_size(self):
    #print "cull_size",self.cachesize
    #_loginfo("CACHE: Culling excessive size (%s)"%self.cachesize)
    while self.cachesize>self.sizelimit:
      self._remove(self._get_oldest())
    #print "done"
   

        
  def _insert(self,key,lifetime,size,data):
    #_loginfo("CACHE: _insert key=%s size=%s lifetime=%s" % (key,size,lifetime))
    if self._exists(key):
      self._remove(key)
    self.cache[key]=(lifetime+time.time(),size,data)
    self.cachesize+=size
    if len(self.cache)>self.itemlimit:
      self._cull_num()
    if self.cachesize>self.sizelimit:
      self._cull_size()
    
  
  def insert(self,key,lifetime,data):
    """
    Insert a new key into the cache, with associated expiry lifetime, size and data. Triggers volume cleaning if the new cache size exceeds the maximum.
    @param key: key for the new data - should be a string
    @param lifetime: desired cache lifetime of this object in seconds
    @param data: free data object. Preferably doesn't refer to other objects in the cache, or size calculations will become impossible.
    """
    self.lock.acquire()
    try:
      try:
        data = cPickle.dumps(data,cPickle.HIGHEST_PROTOCOL)
        self._insert(key,lifetime,len(data),data)
      except cPickle.PickleError:
        _logwarn("CACHE: Pickling error with key=%s"%key)
        pass
    finally:
      self.lock.release()
    
  
  def _remove(self,key):
    #_loginfo("CACHE: _remove key=%s"%key)
    size=-1
    if self._exists(key):
      size = self.cache[key][1]
      self.cachesize-=size
      del self.cache[key]
          
    return size
  
  def remove(self,key):
    """
    Remove a key from the cache, returning the freed space. Returns -1 if the key wasn't found (or if you set the key size to -1, which would be stupid.
    @param key: cache key to remove
    """
    size = -1
    self.lock.acquire()
    try:
      size = self._remove(key)
    finally:
      self.lock.release()
    return size
  
  def retrieve(self,key,default=None):
    #_loginfo("CACHE: Retrieve key=%s"%key)
    """
    Attempts to retrieve the data associated with a key from the cache. Returns default (None) if not found.
    @param key: cache key to retrieve data for
    @param default: what to return if the key is not in the cache (None)
    """
    #print "PC: Getting key [%s]" % (key)
    result=default
    self.lock.acquire()
    try:
      if self._exists(key):
        try:
          result = cPickle.loads(self.cache[key][2])
        except cPickle.PickleError:
          _logwarn("CACHE: Unpickling error with key=%s"%key)
          pass
    finally:
      self.lock.release()
    return result
  
  def _exists(self,key):
    return key in self.cache
  
  def exists(self,key):
    result=False
    self.lock.acquire()
    try:
      result = self._exists(key)
    finally:
      self.lock.release()
    #print "PC: Testing key existence [%s], %s" % (key,result)
    return result
  
  def cachesize(self):
    """
    Get the current "size" of the cache
    """
    self.lock.acquire()
    try:
      result = self.cachesize
    finally:
      self.lock.release()
    return result
    
  def itemcount(self):
    """
    Get the number of items in the cache
    """
    self.lock.acquire()
    try:
      result = len(self.cache)
    finally:
      self.lock.release()
    return result
  
  def keylist(self):
    """
    Get a shallow-copied list of the keys in the cache
    """
    self.lock.acquire()
    try:
      result = self.cache.keys()
    finally:
      self.lock.release()
    return result
    
  def __getitem__(self,key):
    return self.retrieve(key)
  
  def __delitem__(self,key):
    return self.remove(key)
  
  def __len__(self):
    return self.itemcount()
  
  def __iter__(self):
    return self.keylist().__iter__()
  
  def __contains__(self,key):
    return self.exists(key)
  
  def key_event_exists(self,key):
    result = False
    self.lock.acquire()
    try:
      if key in self.cacheevents:
        result = True
      
    finally:
      self.lock.release()
    return result
    
  def key_event_get(self,key):
    event=None
    self.lock.acquire()
    try:
      if key in self.cacheevents:
        event=self.cacheevents[key]
      else:
        event=Event()
        self.cacheevents[key]=event
    finally:
      self.lock.release()
    return event
    #release global lock, acquire local lock
  
  def cacheFetch(self,cachekey,f_produce,*args,**kwargs):
    if cachekey in self:
      #_loginfo("CACHE: cacheFetch (in cache) key=%s"%cachekey)
      cachedata = self[cachekey]
    else:
      #_loginfo("CACHE: cacheFetch (creating) key=%s"%cachekey)
      cachedata,lifetime = f_produce(*args,**kwargs)
      #_loginfo("CACHE: cacheFetch (finished creating) key=%s"%cachekey)
      if lifetime>0:
        self.insert(cachekey,lifetime,cachedata)
    return cachedata
    
      
  def lockedCacheFetch(self,cachekey,f_produce,*args,**kwargs):
    #_loginfo("CACHE: lockedCacheFetch key=%s"%cachekey)
    
    if cachekey in self:
      return self[cachekey]
    else:
      if self.key_event_exists(cachekey):
        event = self.key_event_get(cachekey)
        event.wait(120+random.random()*180) # wait between 2 and 5 minutes before starting regardless
      else:
        event = self.key_event_get(cachekey)
      
      try:
        if cachekey in self:
          cachedata = self[cachekey]
        else:
          cachedata, lifetime = f_produce(*args, **kwargs)
          if lifetime>0:
            self.insert(cachekey,lifetime,cachedata)
        return cachedata
      finally:
        if event:
          event.set()
  
  def asyncWorker(self,cachekey,f_produce,*args,**kwargs):
    """
    Worker function for async threads. This is basically a copy of lockedCacheFetch, that doesn't return data and does extra error checking - since this is usually called by the main thread so exceptions will kill the server.
    """
    if self.key_event_exists(cachekey):
      event = self.key_event_get(cachekey)
      event.wait(120+random.random()*180) # wait between 2 and 5 minutes before starting regardless
    else:
      event = self.key_event_get(cachekey)
      
    try:
      if not cachekey in self:
        try:
          cachedata, lifetime = f_produce(*args, **kwargs)
          if lifetime>0:
            self.insert(cachekey,lifetime,cachedata)
        except:
          pass
    finally:
      if event:
        event.set()
          
  def asyncCacheFill(self,cachekey,f_produce,*args,**kwargs):
    """
    Trigger a thread to try and asynchronously fill some data into the cache. Good for __init__ methods to prefill some information they require before request time.
    """
    kwargs['cachekey']=cachekey
    kwargs['f_produce']=f_produce
    async_fill_thread = Thread(name="asyncCacheFiller:%s"%cachekey,target=self.asyncWorker,args=args,kwargs=kwargs)
    async_fill_thread.start()
      
  def plot(self,*args,**kwargs):
    if args[0]=='list':
      result="<table>\n"
      result+="<tr><td>Key</td><td>Expiry</td><td>Remaining</td><td>Size</td><td>Data</td></tr>\n"
      for key in self.cache:
        result+="<tr>%s</tr>\n"%''.join(['<td>%s</td>'%i for i in [key,self.cache[key][0],self.cache[key][0]-time.time(),self.cache[key][1],'<a href="/overview/cache/data/%s">data</a>'%(key)]])
      result+="</table>"
      return 'text/html',result
    if args[0]=='data':
      key = args[1]
      data = cPickle.dumps(self.cache[key][2])
      if type(data)==type(()):
        result="<html><body>"
        for i in range(len(data)):
          result+="<img src='/overview/cache/item/%s/%s'><hr>"%(key,i)
        result+="</body></html>"
        return 'text/html',result
      else:
        return 'text/plain',repr(data)
    if args[0]=='item':
      key = args[1]
      item = int(args[2])
      return 'image/png',cPickle.dumps(self.cache[key][2])[item]
