/** A hidden internal workspace for making data sample selections.

    DQMSample displays is a temporary internal workspace into which
    user is transferred whenever they request to select a different
    sample.  Once the user selects another sample, or chooses to
    return without making a change, the view is restored to the
    workspace and settings which were interrupted.

    The workspace eliminates most of the header field except to show
    the currently selected data sample name and the run number if
    applicable.  It displays the list of available datasets and runs
    (for data) or CMSSW versions (for relvals) in the canvas area,
    preceded by a search field.  Datasets are grouped by type and
    ordered by the server.

    Change of the text in the search field causes a server side
    incremental search to narrow down the list of choices.  Each
    space-separated token in the search field is taken to be a
    criteria which is matched case-insensitively against the dataset
    name, and if the term is purely a number, against the run number.
    All the terms must match, so for example "cosmics relval" would
    narrow the list down to all datasets which have both "cosmics" and
    "relval" in their names. */
GUI.Plugin.DQMSample = new function() {
  /** Reference to this, for use in callbacks where "this" is
      ambiguous or at least not this closure. */
  var _self = this;

  /** Reference to the GUI main object, for making server calls. */
  var _gui = null;

  /** The DOM element for the canvas area where we display contents. */
  var _canvas = $('canvas');

  /** The DOM element for the main header. */
  var _header = $('header');

  /** The DOM element for the sub-header below the main header, where
      we show the current dataset and run number. */
  var _subheader = $('subheader');

  /** The DOM element for the main search field. */
  var _search = null;

  /** A time-out timer for auto-search.  Started when a key press or
      paste is received on the search field.  Once the timer expires
      will check if the search input field has changed and if so
      issues an incremental server side search. */
  var _searchDelay = null;

  /** The last search input field value seen.  Used by the search
      timer function to determine if a new search is needed. */
  var _searchString = "";

  /** The DOM element for showing the list of available datasets. */
  var _datasets = null;

  /** The DOM element for order by run. */
  var _runorder = null;

  /** The DOM element for order by dataset. */
  var _dsorder = null;

  /** The most recent server state. */
  var _data = null;

  /** Template how to represent a selected element. */
  var _selem = "<em style='color:#c30;background-color:#fff0f0'>";

  // ----------------------------------------------------------------
  /** Attach this plug-in to the GUI display.  Fills the header and
      subheader document elements with choices appropriate to the
      sample selection, and creates the navigation document elements
      on the canvas. */
  this.attach = function(gui)
  {
    _gui = gui;
    _canvas.innerHTML
      = "<div align='center' style='white-space:nowrap;margin:1em 1em'>Search: "
      + "<input id='sample-search' type='text' style='width:50%' value=''"
      + " onkeyup='return GUI.Plugin.DQMSample.startSearch()'"
      + " paste='return GUI.Plugin.DQMSample.startSearch()' /></div>"
      + "<div align='center' style='white-space:nowrap;margin:1em 1em'>"
      + "Vary: <a id='vary-any' href='#' onclick='return GUI.Plugin."
      + "DQMSample.modify(\"vary=any\")'>Any</a>, <a id='vary-run' href='#'"
      + " onclick='return GUI.Plugin.DQMSample.modify(\"vary=run\")'>Run</a>, "
      + "<a id='vary-dataset' href='#' onclick='return GUI.Plugin.DQMSample."
      + "modify(\"vary=dataset\")'>Dataset</a>&nbsp;&nbsp;|&nbsp;&nbsp;"
      + "Order by: <a id='order-by-run' href='#' onclick='return GUI.Plugin."
      + "DQMSample.modify(\"order=run\")'>Run</a>, <a id='order-by-dataset'"
      + " href='#' onclick='return GUI.Plugin.DQMSample.modify(\"order=dataset\")"
      + "'>Dataset</a></div><div align='center' style='margin:1em 1em'>"
      + "<div id='datasets' align='left'></div></div>";
    _search = $('sample-search');
    _datasets = $('datasets');
  };

  /** Detach this plug-in from the GUI display.  Clears the header,
      sub-header and canvas document elements from the objects
      #attach() added and clears closure references for garbage
      collection. */
  this.detach = function()
  {
    // Clear any 'heading' elements we added.
    for (var m, n = _header.firstChild; n; n = m)
    {
      m = n.nextSibling;
      if (n.className == "heading")
	_header.removeChild(n);
    }

    // Clear subheader entirely.
    while (_subheader.firstChild)
      _subheader.removeChild(_subheader.firstChild);

    // Clear the canvas.
    while (_canvas.firstChild)
      _canvas.removeChild(_canvas.firstChild);

    // Turn off searching if it was activated.
    _searchString = "";
    if (_searchDelay)
    {
      clearTimeout(_searchDelay);
      _searchDelay = null;
    }

    _data = null;
    _datasets = null;
  };

  // ----------------------------------------------------------------
  /** Response callback to return without making any changes. */
  this.goback = function()
  {
    _gui.makeCall(_url() + "/return");
    return false;
  };

  /** Response callback to make a dataset selection. */
  this.select = function(type, dsname, run)
  {
    _gui.makeCall(_url() + "/select?type="
		  + encodeURIComponent(type) + ";dataset="
		  + encodeURIComponent(dsname) + ";runnr="
		  + encodeURIComponent(run));
    return false;
  };

  /** Response callback triggered by changes to the search input.
      Triggers the actual search response 50ms after the last
      change has happened. */
  this.startSearch = function()
  {
    if (_searchDelay)
      clearTimeout(_searchDelay);
    _searchDelay = setTimeout(_self.search, 150);
    return true;
  };

  /** Response callback for actually executing a search; user is no
      longer immediately modifying the search field.  Issues a server
      side search request with the current search pattern.

      Note that #update() will refrain from updating the search field
      value if the user has continued to edit the search field between
      the time the request is submitted here and the server responds.
      In that case we'll automatically issue another search shortly
      afterwards. */
  this.search = function()
  {
    if (_searchString != _search.value)
    {
      _searchString = _search.value;
      _gui.makeCall(_url() + "/modify?pat=" + encodeURIComponent(_search.value));
    }
  };

  /** Response callback for changing listing options. */
  this.modify = function(arg)
  {
    _gui.makeCall(_url() + "/modify?" + arg);
    return false;
  };

  // ----------------------------------------------------------------
  /** Update this plug-in display.  Updates the menu areas and the
      canvas navigation area to the current state. */
  this.update = function(data)
  {
    _data = data;

    // Fill in the sub-header area with current sample identification.
    var type = "";
    var cur = data.current;
    var runid = ", Run " + cur.run;
    var dsname = _sanitise(cur.dataset);

    if (cur.type == "live")
      type = "Live", dsname = runid = "";
    else if (cur.type == "online_data")
      type = "Online", dsname = "";
    else if (cur.type == "offline_data")
      type = "Offline: ";
    else if (cur.type == "offline_relval")
      type = "RelVal: ", runid = "";
    else if (cur.type == "offline_mc")
      type = "MC: ", runid = "";
    else
      type = "Other: ";

    var content = "<span class='item'>" + type + dsname + runid
		  + "</span><span class='item'><a href='#' onclick='"
		  + "return GUI.Plugin.DQMSample.goback()'>"
		  + "(Back)</a></span>";

    if (_subheader.innerHTML != content)
      _subheader.innerHTML = content;

    // Update the search value field *iff* the value is different and
    // the user didn't edit the value since the server call.
    if (! _searchDelay && _search.value != data.search)
      _search.value = _searchString = data.search;

    // Update highlight on search criteria.
    $('vary-any').style.fontWeight = (data.vary == "any" ? 'bold' : 'normal');
    $('vary-run').style.fontWeight = (data.vary == "run" ? 'bold' : 'normal');
    $('vary-dataset').style.fontWeight = (data.vary == "dataset" ? 'bold' : 'normal');
    $('order-by-run').style.fontWeight = (data.order == "run" ? 'bold' : 'normal');
    $('order-by-dataset').style.fontWeight = (data.order == "dataset" ? 'bold' : 'normal');

    // Build the dataset area contents and update if changed.
    content = this.display(cur, data.items, data.order);
    if (_datasets.innerHTML != content)
      _datasets.innerHTML = content;
  };

  this.display = function(cur, items, order)
  {
    // Build the dataset area listing, grouping datasets by sample
    // category, then sample name (folding similar relval names), and
    // then by run (for data) or CMSSW version (for relval).  Note
    // that simulated and live samples will have restricted choices.
    var content = "";
    var byrun = (order == "run");
    for (var i = 0, e = items.length; i < e; ++i)
    {
      var g = items[i];
      var sel = (g.type == cur.type);
      content += "<div style='clear:both;font-weight:bold'>" + (sel ? _selem : "");

      if (g.type == "live")
	content += "<a href='#' onclick='return GUI.Plugin.DQMSample.select(\""
		   + g.type + "\", \"" + g.items[0].dataset + "\", "
		   + g.items[0].run + ")'>Live</a>";
      else if (g.type == "online_data")
	content += "Online";
      else if (g.type == "offline_data")
	content += "Offline";
      else if (g.type == "offline_relval")
	content += "RelVal";
      else if (g.type == "offline_mc")
	content += "MC";
      else
	content += "Other";

      content += (sel ? "</em>" : "") + "</div>";

      if (order == "run")
        content += this.byrun(cur, g.items, g.type);
      else
        content += this.bydataset(cur, g.items, g.type);
    }

    return content;
  };

  this.bydataset = function(cur, items, type)
  {
    var content = "";
    var prev = null;
    var template = "<a href='#' style='display:block;float:left;"
		   + "margin-right:1em;min-width:5em' onclick='return"
		   + " GUI.Plugin.DQMSample.select"
		   + "(\"${type}\", \"${dsname}\", ${run})'>"
		   + "${begin}${val}${end}</a>";

    for (var j = 0, k = items.length; j < k; ++j)
    {
      var s = items[j];
      var dataset = s.dataset;
      if (type == "relval")
        dataset = dataset.replace(/CMSSW(_\d+)+(_pre\d+)?/, "CMSSW_x_y_z");
      var selds = (cur.type == s.type && cur.dataset == s.dataset)
      var selrun = (selds && cur.run == s.run);
      var selver = (selds && cur.version == s.version);

      if (prev != null && dataset != prev)
	content += "</div>";

      if (prev == null || dataset != prev)
      {
	prev = dataset;
	if (type == "live")
	  prev = null;
	else if (type == "online_data")
	  content += "<div style='clear:both;margin:0 2em'>";
	else if (s.run > 0 || s.version != "")
	  content += _("<div style='clear:both;margin:0 2em'>${begin}${dshtml}${end}:<br />",
	               { dshtml: _sanitise(dataset),
			 begin: (selds ? _selem : ""),
			 end: (selds ? "</em>" : "") });
	else
	  content += _("<div style='clear:both;margin:0 2em'>${begin}<a href='#' onclick='return"
		       + " GUI.Plugin.DQMSample.select(\"${type}\",\"${dsname}\",${run})'>"
		       + "${dshtml}</a>${end}",
	               { type: s.type, run: s.run, dsname: s.dataset,
			 dshtml: _sanitise(dataset),
			 begin: (selds ? _selem : ""),
			 end: (selds ? "</em>" : "") });

      }

      if (s.run > 0)
	content += _(template,
	    	     { type: s.type, run: s.run, dsname: s.dataset, val: s.run,
		       begin: (selrun ? _selem : ""), end: (selrun ? "</em>" : "") });
      else if (s.version != "")
	content += _(template,
	    	     { type: s.type, run: s.run, dsname: s.dataset,
		       val: s.version.replace(/^CMSSW_/, ""),
		       begin: (selver ? _selem : ""), end: (selver ? "</em>" : "") });
    }

    if (prev != null)
      content += "</div>";

    return content;
  };

  this.byrun = function(cur, items, type)
  {
    var content = "";
    var prev = null;
    var template
      = (type == "live" ? ""
	 : (type == "online_data"
	    ? "<a href='#' style='display:block;float:left;"
	      + "margin-right:1em;min-width:5em' onclick='return GUI.Plugin"
	      + ".DQMSample.select(\"${type}\", \"${dsname}\", ${run})'>"
	      + "${begin}${val}${end}</a>"
	    : "<div style='margin-left:1em'><a href='#'"
	      + " onclick='return GUI.Plugin.DQMSample.select"
	      + "(\"${type}\", \"${dsname}\", ${run})'>${begin}${val}${end}"
	      + "</a></div>"));

    for (var j = 0, k = items.length; j < k; ++j)
    {
      var s = items[j];
      var key = (type == "offline_data" ? s.run
		 : (type == "offline_relval" ? s.version : type));
      var selrun = (cur.type == s.type && cur.run == s.run);
      var selver = (cur.type == s.type && cur.version == s.version);
      var selds = (cur.type == s.type
		   && cur.dataset == s.dataset
		   && ((s.run > 0 && selrun)
		       || (s.version != "" && selver)
		       || (s.run == 0 && s.version == "")));
      var opts = { type: s.type, run: s.run, dsname: s.dataset,
		   val: (type.indexOf("offline_") >= 0 ? _sanitise(s.dataset) : s.run),
		   begin: (selds ? _selem : ""),
		   end: (selds ? "</em>" : "") };

      if (prev != null && key != prev)
	content += "</div>";

      if (prev == null || key != prev)
      {
	prev = key;
	if (type == "live")
	  prev = null;
	else if (type == "online_data" || type == "offline_mc")
	  content += "<div style='clear:both;margin:0 2em'>";
	else
	  content += _("<div style='clear:both;margin:0 2em'>${begin}${key}${end}:<br />",
	               { key: _sanitise(key),
			 begin: (selrun && selver ? _selem : ""),
			 end: (selrun && selver ? "</em>" : "") });
      }

      content += _(template, opts);
    }

    if (prev != null)
      content += "</div>";

    return content;
  };

  return this;
}();
